from mads_datasets.datasets.basicdatasets import (
    FacesDataset,
    PdDataset,
    PolarsDataset,
    TextDataset,
)
from mads_datasets.datasets.torchdatasets import (
    ImgDataset,
    MNISTDataset,
    SunspotDataset,
    TensorDataset,
    TSDataset,
)

__all__ = [
    "PdDataset",
    "PolarsDataset",
    "FacesDataset",
    "TextDataset",
    "SunspotDataset",
    "TensorDataset",
    "MNISTDataset",
    "ImgDataset",
    "TSDataset",
]



from mads_datasets.factories.basicfactories import (
    FavoritaDatasetFactory,
    IMDBDatasetFactory,
    IrisDatasetFactory,
    PenguinsDatasetFactory,
    SecureDatasetFactory,
)
from mads_datasets.factories.torchfactories import (
    FashionDatasetFactory,
    FlowersDatasetFactory,
    GesturesDatasetFactory,
    SunspotsDatasetFactory,
)

__all__ = [
    "PenguinsDatasetFactory",
    "FavoritaDatasetFactory",
    "IrisDatasetFactory",
    "IMDBDatasetFactory",
    "SecureDatasetFactory",
    "SunspotsDatasetFactory",
    "FashionDatasetFactory",
    "GesturesDatasetFactory",
    "FlowersDatasetFactory",
]



import re
import string
from enum import Enum
from pathlib import Path
from typing import Callable, List, Optional, Tuple, cast

from pydantic import BaseModel, HttpUrl


class FileTypes(Enum):
    JPG = ".jpg"
    PNG = ".png"
    TXT = ".txt"
    ZIP = ".zip"
    TGZ = ".tgz"
    TAR = ".tar.gz"
    GZ = ".gz"
    PT = ".pt"
    CSV = ".csv"
    PARQ = ".parq"


class ReportTypes(Enum):
    GIN = 1
    TENSORBOARD = 2
    MLFLOW = 3
    RAY = 4


class DatasetType(Enum):
    FLOWERS = 1
    IMDB = 2
    GESTURES = 3
    FASHION = 4
    SUNSPOTS = 5
    IRIS = 6
    PENGUINS = 7
    FAVORITA = 8
    SECURE = 9


class BaseSettings(BaseModel):
    def __str__(self) -> str:
        return "\n".join(f"{k}: {v}" for k, v in self.__dict__.items())

    def __repr__(self) -> str:
        return "\n".join(f"{k}: {v}" for k, v in self.__dict__.items())


class DatasetSettings(BaseSettings):
    dataset_url: HttpUrl
    filename: Path
    name: str
    unzip: bool
    formats: Optional[List[FileTypes]] = None
    digest: Optional[str] = None


class ImgDatasetSettings(DatasetSettings):
    trainfrac: float
    img_size: Tuple[int, int]


class TextDatasetSettings(DatasetSettings):
    maxvocab: int
    maxtokens: int
    clean_fn: Callable


class WindowedDatasetSettings(DatasetSettings):
    horizon: int
    window_size: int


class PdDatasetSettings(DatasetSettings):
    target: str
    features: List[str]


class SecureDatasetSettings(DatasetSettings):
    keyaccount: str
    keyname: str


favoritasettings = DatasetSettings(
    dataset_url=cast(
        HttpUrl, "https://gitlab.com/api/v4/projects/50444079/repository/files/favorita%2Ezip/raw?lfs=true"
    ),
    filename=Path("favorita.zip"),
    name="favorita",
    unzip=True,
    formats=[FileTypes.PARQ],
    digest="284fa8134fce7d918d95ccd975c8f14d"
)


penguinssettings = DatasetSettings(
    dataset_url=cast(
        HttpUrl,
        "https://github.com/raoulg/juliaintro/raw/main/data/palmerpenguins.parq",
    ),
    filename=Path("penguins.parq"),
    name="penguins",
    unzip=False,
    digest="675e2a75750d0e076df810893e675476",
)

irissettings = PdDatasetSettings(
    dataset_url=cast(
        HttpUrl, "https://github.com/raoulg/data_assets/raw/main/iris_dirty.csv"
    ),
    filename=Path("iris.csv"),
    name="iris",
    unzip=False,
    target="class",
    features=["sepal_length", "sepal_width", "petal_length", "petal_width"],
    digest="3679279dc61f6fdd859be9888db27f75",
)

sunspotsettings = WindowedDatasetSettings(
    dataset_url=cast(HttpUrl, "https://www.sidc.be/SILSO/DATA/SN_m_tot_V2.0.txt"),
    filename=Path("sunspots.txt"),
    name="sunspots",
    unzip=False,
    formats=[],
    horizon=3,
    window_size=26,
    digest="b0c58f706c1e14a1faaa35635b51679e",
)

fashionmnistsettings = DatasetSettings(
    dataset_url=cast(
        HttpUrl, "https://github.com/raoulg/data_assets/raw/main/fashionmnist.pt"
    ),
    filename=Path("fashionmnist.pt"),
    name="fashionmnist",
    unzip=False,
    formats=[FileTypes.PT],
    digest="c4f1c3f76673fe3802f579773267163a",
)

gesturesdatasetsettings = DatasetSettings(
    dataset_url=cast(
        HttpUrl, "https://github.com/raoulg/gestures/raw/main/gestures-dataset.zip"
    ),
    filename=Path("gestures-dataset.zip"),
    name="gestures",
    unzip=True,
    formats=[FileTypes.TXT],
    digest="7966323b95154f314e831c312e5cc33b",
)


def clean(text: str) -> str:
    punctuation = f"[{string.punctuation}]"
    # remove CaPiTaLs
    lowercase = text.lower()
    # change don't and isn't into dont and isnt
    neg = re.sub("\\'", "", lowercase)
    # swap html tags for spaces
    html = re.sub("<br />", " ", neg)
    # swap punctuation for spaces
    stripped = re.sub(punctuation, " ", html)
    # remove extra spaces
    spaces = re.sub("  +", " ", stripped)
    return spaces


imdbdatasetsettings = TextDatasetSettings(
    dataset_url=cast(
        HttpUrl, "https://ai.stanford.edu/~amaas/data/sentiment/aclImdb_v1.tar.gz"
    ),
    filename=Path("aclImdb_v1.tar.gz"),
    name="imdb",
    unzip=True,
    formats=[FileTypes.TXT],
    maxvocab=10000,
    maxtokens=100,
    clean_fn=clean,
    digest="7c2ac02c03563afcf9b574c7e56c153a",
)

flowersdatasetsettings = ImgDatasetSettings(
    dataset_url=cast(
        HttpUrl,
        "https://storage.googleapis.com/download.tensorflow.org/example_images/flower_photos.tgz",
    ),
    filename=Path("flowers.tgz"),
    name="flowers",
    unzip=True,
    formats=[FileTypes.JPG],
    trainfrac=0.8,
    img_size=(224, 224),
    digest="6f87fb78e9cc9ab41eff2015b380011d",
)



from pathlib import Path
from loguru import logger

from mads_datasets.base import AbstractDatasetFactory
from mads_datasets.factories import (
    FashionDatasetFactory,
    FavoritaDatasetFactory,
    FlowersDatasetFactory,
    GesturesDatasetFactory,
    IMDBDatasetFactory,
    IrisDatasetFactory,
    PenguinsDatasetFactory,
    SecureDatasetFactory,
    SunspotsDatasetFactory,
)
from mads_datasets.settings import (
    DatasetType,
    fashionmnistsettings,
    favoritasettings,
    flowersdatasetsettings,
    gesturesdatasetsettings,
    imdbdatasetsettings,
    irissettings,
    penguinssettings,
    sunspotsettings,
    SecureDatasetSettings,
)

__all__ = ["DatasetFactoryProvider", "DatasetType"]

__version__ = "0.3.10"


class DatasetFactoryProvider:
    @staticmethod
    def create_factory(dataset_type: DatasetType, **kwargs) -> AbstractDatasetFactory:
        datadir = Path(kwargs.get("datadir", Path.home() / ".cache/mads_datasets"))
        if dataset_type == DatasetType.FLOWERS:
            return FlowersDatasetFactory(flowersdatasetsettings, datadir=datadir)
        if dataset_type == DatasetType.IMDB:
            return IMDBDatasetFactory(imdbdatasetsettings, datadir=datadir)
        if dataset_type == DatasetType.GESTURES:
            return GesturesDatasetFactory(gesturesdatasetsettings, datadir=datadir)
        if dataset_type == DatasetType.FASHION:
            return FashionDatasetFactory(fashionmnistsettings, datadir=datadir)
        if dataset_type == DatasetType.SUNSPOTS:
            return SunspotsDatasetFactory(sunspotsettings, datadir=datadir)
        if dataset_type == DatasetType.IRIS:
            return IrisDatasetFactory(irissettings, datadir=datadir)
        if dataset_type == DatasetType.PENGUINS:
            return PenguinsDatasetFactory(penguinssettings, datadir=datadir)
        if dataset_type == DatasetType.FAVORITA:
            return FavoritaDatasetFactory(favoritasettings, datadir=datadir)
        if dataset_type == DatasetType.SECURE:
            securesettings = kwargs.get("settings", None)
            if not securesettings:
                logger.warning("No settings provided for SecureDatasetFactory.")
            if not isinstance(securesettings, SecureDatasetSettings):
                raise ValueError(
                    f"Invalid settings type: {type(securesettings)}. "
                    "Expected SecureDatasetSettings."
                )
            return SecureDatasetFactory(securesettings, datadir=datadir)

        raise ValueError(f"Invalid dataset type: {dataset_type}")



from typing import List, Tuple
from pathlib import Path
import pytest
from mads_datasets.datatools import iter_valid_paths, get_file
from mads_datasets.settings import FileTypes
import responses
from unittest import mock

class TestData:
    @pytest.fixture
    def mock_directory(self, tmpdir_factory) -> Tuple[Path, List[str], List[str]]:
        dir = Path(tmpdir_factory.mktemp("data").strpath)
        formats = [".jpg", ".png"]
        classes = ["class1", "class2"]
        for cls in classes:
            for i in range(2):
                for format_ in formats:
                    file = dir / cls / f"image{i}{format_}"
                    file.parent.mkdir(parents=True, exist_ok=True)
                    with open(file, "w") as f:
                        f.write("test")
        return dir, classes, formats

    def test_iter_valid_paths(self, mock_directory):
        path, expected_classes, formats = mock_directory
        paths, classes = iter_valid_paths(Path(path), [FileTypes.JPG, FileTypes.PNG])

        # Check if classes are correctly recognized
        assert sorted(classes) == sorted(expected_classes)

        # Check if all paths are returned and they have correct formats
        returned_files = list(paths)
        assert len(returned_files) == len(expected_classes) * 2 * len(formats)
        for path in returned_files:
            assert path.suffix in formats

        # Check if files belong to correct directories (classes)
        for path in returned_files:
            assert path.parent.name in expected_classes


@responses.activate
def test_get_file(tmpdir):
    # Setup
    data_dir = Path(tmpdir)
    filename = "file.txt"
    url = "http://example.com/"
    path = data_dir / filename

    responses.add(responses.GET, url, body=b"content")

    with mock.patch.object(Path, "exists", return_value=False):
        with mock.patch("builtins.open", mock.mock_open()) as mock_file:
            with mock.patch("mads_datasets.datatools.extract") as mock_extract:
                # Call the function
                result = get_file(data_dir, filename, url, unzip=True, overwrite=False)

    # Check the results
    assert result == path
    mock_file.assert_called_once_with(path, "wb")
    mock_extract.assert_called_once_with(path)

    assert len(responses.calls) == 1
    assert responses.calls[0].request.url == url

@responses.activate
def test_get_file_already_exists(tmpdir):
    # Setup
    data_dir = Path(tmpdir)
    filename = "file.txt"
    url = "http://example.com"
    path = data_dir / filename

    responses.add(responses.GET, url, body=b"content")

    with mock.patch.object(Path, "exists", return_value=True):
        # Call the function
        result = get_file(data_dir, filename, url, unzip=True, overwrite=False)

    # Check the results
    assert result == path
    assert len(responses.calls) == 0  # The file already exists, so no request should be made



from pathlib import Path
from typing import TYPE_CHECKING, List, Tuple

from PIL import Image
from tqdm import tqdm

from mads_datasets.base import AbstractDataset, DatasetProtocol

if TYPE_CHECKING:
    import pandas as pd
    import polars as pl


class PdDataset(DatasetProtocol):
    def __init__(
        self,
        df: "pd.DataFrame",  # noqa: F821 type: ignore
        target: str,
        features: List[str],  # noqa: F821 type: ignore
    ) -> None:  # noqa: F821
        self.df = df
        self.target = target
        self.features = features

    def __len__(self) -> int:
        return len(self.df)

    def __getitem__(
        self, idx: int
    ) -> Tuple["pd.Series", "pd.Series"]:  # noqa: F821 type: ignore
        x = self.df[self.features].iloc[idx]
        y = self.df[self.target].iloc[idx]
        return x, y

    def __repr__(self) -> str:
        return f"PdDataset (len {len(self)})"


class PolarsDataset(DatasetProtocol):
    def __init__(self, df: "pl.DataFrame"):
        self.df = df

    def __len__(self) -> int:
        return len(self.df)

    def __getitem__(self, idx) -> "pl.DataFrame":
        return self.df[idx]


class FacesDataset(AbstractDataset):
    def __init__(self, paths: List[Path]) -> None:
        super().__init__(paths)

    def process_data(self) -> None:
        for path in self.paths:
            img = self.load_image(path)
            self.dataset.append((img, path.name))

    def load_image(self, path: Path) -> Image.Image:
        img = Image.open(path)
        return img


class TextDataset(AbstractDataset):
    """This assumes textual data, one line per file"""

    def process_data(self) -> None:
        for file in tqdm(self.paths, colour="#1e4706"):
            with open(file) as f:
                x = f.readline()
            y = file.parent.name
            self.dataset.append((x, y))

    def __repr__(self) -> str:
        return f"TextDataset (len {len(self)})"



import random
import shutil
from copy import deepcopy
from typing import TYPE_CHECKING, Any, Mapping

import numpy as np

from mads_datasets.base import AbstractDatasetFactory, DatasetProtocol
from mads_datasets.datasets import ImgDataset, MNISTDataset, SunspotDataset, TSDataset
from mads_datasets.datatools import (
    import_torch,
    iter_valid_paths,
    keep_subdirs_only,
    walk_dir,
    window,
)
from mads_datasets.settings import (
    DatasetSettings,
    ImgDatasetSettings,
    WindowedDatasetSettings,
)

if TYPE_CHECKING:
    import torch

    Tensor = torch.Tensor


class SunspotsDatasetFactory(AbstractDatasetFactory[WindowedDatasetSettings]):
    """
    Data from https://www.sidc.be/SILSO/datafiles
    """

    def create_dataset(
        self, *args: Any, **kwargs: Any
    ) -> Mapping[str, DatasetProtocol]:
        self.download_data()
        with import_torch() as torch:  # type: ignore
            spots = np.genfromtxt(str(self.filepath), usecols=(3))  # type: ignore
            tensors = torch.from_numpy(spots).type(torch.float32)  # type: ignore

            split = kwargs.pop("split", 0.8)
            idx = int(len(tensors) * split)
            train = tensors[:idx]
            valid = tensors[idx:]

            norm = max(train)
            train = train / norm
            valid = valid / norm

            window_size = self.settings.window_size
            horizon = self.settings.horizon
            trainset = SunspotDataset(self._window(train, window_size), horizon)
            validset = SunspotDataset(self._window(valid, window_size), horizon)
            return {"train": trainset, "valid": validset}

    @staticmethod
    def _window(data: "Tensor", window_size: int) -> "Tensor":
        idx = window(data, window_size)
        dataset = data[idx]
        dataset = dataset[..., None]
        return dataset


class FashionDatasetFactory(AbstractDatasetFactory[DatasetSettings]):
    def create_dataset(
        self, *args: Any, **kwargs: Any
    ) -> Mapping[str, DatasetProtocol]:
        self.download_data()
        with import_torch() as torch:  # type: ignore
            data = torch.load(self.filepath)  # type: ignore
            training_data = data["traindata"]
            training_labels = data["trainlabels"]
            test_data = data["testdata"]
            test_labels = data["testlabels"]

            train = MNISTDataset(training_data, training_labels)
            test = MNISTDataset(test_data, test_labels)

            return {"train": train, "valid": test}


class GesturesDatasetFactory(AbstractDatasetFactory[DatasetSettings]):
    def __init__(
        self,
        settings: DatasetSettings,
        **kwargs: Any,
    ) -> None:
        super().__init__(settings, **kwargs)
        self._created = False
        self.datasets: Mapping[str, DatasetProtocol]

    def create_dataset(
        self, *args: Any, **kwargs: Any
    ) -> Mapping[str, DatasetProtocol]:
        self.download_data()

        if self._created:
            return deepcopy(self.datasets)
        formats = [f.value for f in self._settings.formats]
        datadir = self.subfolder / "gestures-dataset"
        img = datadir / "gestures.png"
        if img.exists():
            shutil.move(img, datadir.parent / "gestures.png")
        keep_subdirs_only(datadir)
        paths = [path for path in walk_dir(datadir) if path.suffix in formats]
        random.shuffle(paths)

        split = kwargs.pop("split", 0.8)
        idx = int(len(paths) * split)
        trainpaths = paths[:idx]
        validpaths = paths[idx:]

        traindataset = TSDataset(trainpaths)
        validdataset = TSDataset(validpaths)
        datasets = {
            "train": traindataset,
            "valid": validdataset,
        }
        self.datasets = datasets  # type: ignore
        self._created = True
        return datasets


class FlowersDatasetFactory(AbstractDatasetFactory[ImgDatasetSettings]):
    def create_dataset(
        self, *args: Any, **kwargs: Any
    ) -> Mapping[str, DatasetProtocol]:
        self.download_data()
        formats = self._settings.formats
        paths_, class_names = iter_valid_paths(
            self.subfolder / "flower_photos", formats=formats
        )
        paths = [*paths_]
        random.shuffle([paths])
        trainidx = int(len(paths) * self._settings.trainfrac)
        train = paths[:trainidx]
        valid = paths[trainidx:]
        traindataset = ImgDataset(train, class_names, img_size=self._settings.img_size)
        validdataset = ImgDataset(valid, class_names, img_size=self._settings.img_size)
        return {"train": traindataset, "valid": validdataset}



from __future__ import annotations

import shutil
import tarfile
import zipfile
from contextlib import contextmanager
from datetime import datetime
from pathlib import Path
from typing import TYPE_CHECKING, Iterator, List, Optional, Tuple, Union
from keyring.errors import NoKeyringError

import keyring
import requests
import os
from loguru import logger
from tqdm import tqdm

from mads_datasets.settings import FileTypes, SecureDatasetSettings

if TYPE_CHECKING:
    import torch

    Tensor = torch.Tensor


def walk_dir(path: Path) -> Iterator:
    """loops recursively through a folder

    Args:
        path (Path): folder to loop trough. If a directory
            is encountered, loop through that recursively.

    Yields:
        Generator: all paths in a folder and subdirs.
    """

    for p in Path(path).iterdir():
        if p.is_dir():
            yield from walk_dir(p)
            continue
        # resolve works like .absolute(), but it removes the "../.." parts
        # of the location, so it is cleaner
        yield p.resolve()


def iter_valid_paths(
    path: Path, formats: List[FileTypes]
) -> Tuple[Iterator, List[str]]:
    """
    Gets all paths in folders and subfolders
    strips the classnames assuming that the subfolders are the classnames
    Keeps only paths with the right suffix


    Args:
        path (Path): image folder
        formats (List[str]): suffices to keep, eg [".jpg", ".png"]

    Returns:
        Tuple[Iterator, List[str]]: _description_
    """
    # gets all files in folder and subfolders
    walk = walk_dir(path)
    # retrieves foldernames as classnames
    class_names = [subdir.name for subdir in path.iterdir() if subdir.is_dir()]
    # keeps only specified formats
    formats_ = [f.value for f in formats]
    paths = (path for path in walk if path.suffix in formats_)
    return paths, class_names


def get_file(
    data_dir: Path,
    filename: Path,
    url: str,
    unzip: bool = True,
    overwrite: bool = False,
    headers: dict = None,
) -> Path:
    """Download a file from url to location data_dir / filename

    Args:
        data_dir (Path): dir to store file
        filename (Path): filename
        url (str): url to obtain filename
        unzip (bool, optional): If the file needs unzipping
        overwrite (bool, optional): overwrite file, if it already exists.

    Returns:
        Path: _description_
    """

    path = data_dir / filename
    if path.exists() and not overwrite:
        logger.info(f"File {path} already exists, skip download")
        return path
    response = requests.get(url, stream=True, headers=headers)
    if response.status_code != 200:
        raise ValueError(f"Request failed with status code {response.status_code}")
    total_size_in_bytes = int(response.headers.get("content-length", 0))
    block_size = 2**10
    progress_bar = tqdm(
        total=total_size_in_bytes, unit="iB", unit_scale=True, colour="#1e4706"
    )
    logger.info(f"Downloading {path}")
    with open(path, "wb") as file:
        for data in response.iter_content(block_size):
            progress_bar.update(len(data))
            file.write(data)
    progress_bar.close()
    if unzip:
        extract(path)
    return path


def extract(path: Path) -> None:
    try:
        file_type = FileTypes(path.suffix)
    except ValueError:
        logger.info(f"The suffix {path.suffix} is not a recognized file type.")

    if file_type in [FileTypes.ZIP]:
        logger.info(f"Unzipping {path}")
        with zipfile.ZipFile(path, "r") as zip_ref:
            zip_ref.extractall(path.parent)

    if file_type in [FileTypes.TGZ, FileTypes.TAR, FileTypes.GZ]:
        logger.info(f"Unzipping {path}")
        with tarfile.open(path, "r:gz") as tar:
            tar.extractall(path=path.parent)


def clean_dir(dir: Union[str, Path]) -> None:
    dir = Path(dir)
    if dir.exists():
        logger.info(f"Clean out {dir}")
        shutil.rmtree(dir)
    else:
        dir.mkdir(parents=True)


def window(x: "Tensor", n_time: int) -> "Tensor":
    """
    Generates and index that can be used to window a timeseries.
    E.g. the single series [0, 1, 2, 3, 4, 5] can be windowed into 4 timeseries with
    length 3 like this:

    [0, 1, 2]
    [1, 2, 3]
    [2, 3, 4]
    [3, 4, 5]

    We now can feed 4 different timeseries into the model, instead of 1, all
    with the same length.
    """
    with import_torch() as torch:  # type: ignore
        n_window = len(x) - n_time + 1
        time = torch.arange(0, n_time).reshape(1, -1)  # type: ignore
        window = torch.arange(0, n_window).reshape(-1, 1)  # type: ignore
        idx = time + window
        return idx


def dir_add_timestamp(log_dir: Optional[Path] = None) -> Path:
    if log_dir is None:
        log_dir = Path(".")
    log_dir = Path(log_dir)
    timestamp = datetime.now().strftime("%Y%m%d-%H%M")
    log_dir = log_dir / timestamp
    logger.info(f"Logging to {log_dir}")
    if not log_dir.exists():
        log_dir.mkdir(parents=True)
    return log_dir


def keep_subdirs_only(path: Path) -> None:
    files = [file for file in path.iterdir() if file.is_file()]
    for file in files:
        file.unlink()


@contextmanager
def import_torch() -> Optional["torch"]:  # type: ignore
    try:
        import torch
    except ImportError:
        logger.warning(
            "Torch is not installed."
            "Please install 'mads-datasets[torch]' to use this feature."
        )
        yield None
    else:
        yield torch


def check_token(account: str, name: str) -> str:
    haskeyring = True
    notstored = False
    try:
        token = keyring.get_password(account, name)
        if not token:
            notstored = True
    except NoKeyringError as e:
        haskeyring = False
        logger.warning(
            f"Failed to find a keyring backend: {e}"
            f"Falling back to environment to obtain token from variable {account}."
        )
        token = os.environ.get(account)
        if not token:
            logger.warning(
                f"Could not obtain {account} from the environment."
                "Please add this to your environment"
            )

    if not token:
            logger.info("Enter your token for {account} and {name} manually:")
            token = input("Token: ")
    if haskeyring and notstored:
        logger.info(f"Storing token for {account} and {name} in keyring.")
        keyring.set_password(account, name, token)
    return token


def create_headers(settings: SecureDatasetSettings) -> dict:
    token = check_token(settings.keyaccount, settings.keyname)

    headers = {
        "PRIVATE-TOKEN": token,
    }
    return headers



from pathlib import Path
from typing import TYPE_CHECKING, List, Tuple

import numpy as np
from PIL import Image
from tqdm import tqdm

from mads_datasets.base import AbstractDataset, DatasetProtocol
from mads_datasets.datatools import import_torch

if TYPE_CHECKING:
    import torch

    Tensor = torch.Tensor


class SunspotDataset(DatasetProtocol):
    def __init__(self, data: "Tensor", horizon: int) -> None:
        self.data = data
        self.size = len(data)
        self.horizon = horizon

    def __len__(self) -> int:
        return self.size

    def __getitem__(self, idx: int) -> Tuple["Tensor", "Tensor"]:
        # get a single item
        item = self.data[idx]
        # slice off the horizon
        x = item[: -self.horizon, :]
        y = item[-self.horizon :, :].squeeze(
            -1
        )  # squeeze will remove the last dimension if possible.
        return x, y

    def __repr__(self) -> str:
        return f"SunspotDataset (len {len(self)})"


class TensorDataset(DatasetProtocol):
    """The main responsibility of the Dataset class is to
    offer a __len__ method and a __getitem__ method
    """

    def __init__(self, data: "Tensor", targets: "Tensor") -> None:
        self.data = data
        self.targets = targets
        assert len(data) == len(targets)

    def __repr__(self) -> str:
        return f"TensorDataset (len {len(self)})"


class MNISTDataset(DatasetProtocol):
    """MNIST dataset
    Args:
        data (torch.Tensor): images
        labels (torch.Tensor): labels
        transform (callable, optional): Optional transform to be applied
            on a sample.

    Returns:
        torch.Tensor: image
        torch.Tensor: label
    """

    def __init__(self, data: "Tensor", labels: "Tensor"):
        self.data = data
        self.labels = labels

    def __len__(self) -> int:
        return len(self.labels)

    def __getitem__(self, idx) -> Tuple:
        # add single batch dimension
        with import_torch() as torch:  # type: ignore
            image = self.data[idx].unsqueeze(0)
            # scale to [0, 1] and cast to torch.float32
            image = image / 255.0
            image = image.type(torch.float32)  # type: ignore
            label = self.labels[idx].type(torch.uint8)  # type: ignore
            return image, label

    def __repr__(self) -> str:
        return f"MNISTDataset (len {len(self)})"


class ImgDataset(AbstractDataset):
    def __init__(
        self, paths: List[Path], class_names: List[str], img_size: Tuple[int, int]
    ) -> None:
        self.img_size = img_size
        self.class_names = class_names
        super().__init__(paths)

    def process_data(self) -> None:
        with import_torch() as torch:  # type: ignore
            for file in self.paths:
                img = self.load_image(file, self.img_size)
                x_ = np.transpose(img, (2, 0, 1))
                x = torch.tensor(x_ / 255.0).type(torch.float32)  # type: ignore
                y = torch.tensor(self.class_names.index(file.parent.name))
                self.dataset.append((x, y))

    def load_image(self, path: Path, image_size: Tuple[int, int]) -> np.ndarray:
        # load file
        img_ = Image.open(path).resize(image_size, Image.LANCZOS)
        return np.asarray(img_)

    def __repr__(self) -> str:
        return f"ImgDataset (imgsize {self.img_size}, #classes {len(self.class_names)})"


class TSDataset(AbstractDataset):
    """This assume a txt file with numeric data
    Dropping the first columns is hardcoded
    y label is name-1, because the names start with 1

    """

    def process_data(self) -> None:
        with import_torch() as torch:  # type: ignore
            for file in tqdm(self.paths, colour="#1e4706"):
                x_ = np.genfromtxt(file)[:, 3:]
                x = torch.tensor(x_).type(torch.float32)  # type: ignore
                y = torch.tensor(int(file.parent.name) - 1)  # type: ignore
                self.dataset.append((x, y))

    def __repr__(self) -> str:
        return f"TSDataset (size {len(self)})"



import hashlib
import random
from abc import ABC, abstractmethod
from pathlib import Path
from typing import (
    Any,
    Callable,
    Generic,
    Iterator,
    List,
    Mapping,
    Optional,
    Protocol,
    Sequence,
    Tuple,
    TypeVar,
)

import numpy as np
from loguru import logger

from mads_datasets.datatools import create_headers, get_file
from mads_datasets.settings import DatasetSettings, SecureDatasetSettings


class DatasetProtocol(Protocol):
    def __len__(self) -> int:
        ...

    def __getitem__(self, idx: int) -> Any:
        ...


class ProcessingDatasetProtocol(DatasetProtocol):
    def process_data(self) -> None:
        ...


class DatastreamerProtocol(Protocol):
    def stream(self) -> Iterator:
        ...


class AbstractDataset(ABC, ProcessingDatasetProtocol):
    """The main responsibility of the Dataset class is to load the data from disk
    and to offer a __len__ method and a __getitem__ method
    """

    def __init__(self, paths: List[Path]) -> None:
        self.paths = paths
        random.shuffle(self.paths)
        self.dataset: List = []
        self.process_data()

    def __len__(self) -> int:
        return len(self.dataset)

    def __getitem__(self, idx: int) -> Tuple:
        return self.dataset[idx]

    @abstractmethod
    def process_data(self) -> None:
        raise NotImplementedError


T = TypeVar("T", bound=DatasetSettings)


class AbstractDatasetFactory(ABC, Generic[T]):
    def __init__(self, settings: T, datadir: Path) -> None:
        self._settings = settings
        self.datadir = datadir

        if type(settings) == SecureDatasetSettings:
            self.secure = True
        else:
            self.secure = False

    @property
    def settings(self) -> T:
        return self._settings

    def download_data(self) -> None:
        url = self._settings.dataset_url
        filename = self._settings.filename
        self.subfolder = Path(self.datadir) / self.settings.name

        if self.secure:
            headers = create_headers(self._settings)  # type: ignore
        else:
            headers = None

        if not self.subfolder.exists():
            logger.info("Start download...")
            self.subfolder.mkdir(parents=True)
            self.filepath = get_file(
                self.subfolder,
                filename,
                url=str(url),
                unzip=self._settings.unzip,
                overwrite=False,
                headers=headers,
            )
            digest = self.calculate_md5(self.filepath)
            if self.settings.digest is not None:
                if digest != self.settings.digest:
                    raise ValueError(
                        f"Digest of {self.filepath} does not match expected digest"
                        f"\nExpected: {self.settings.digest}\nGot: {digest}"
                    )
                else:
                    logger.info(f"Digest of {self.filepath} matches expected digest")
            else:
                logger.info(f"Digest of downloaded {self.filepath} is {digest}")

            if self._settings.unzip:
                logger.info(f"Removing unzipped file {self.filepath}")
                self.filepath.unlink()

        else:
            logger.info(f"Folder already exists at {self.subfolder}")
            self.filepath = self.subfolder / filename
            if self.filepath.exists():
                logger.info(f"File already exists at {self.filepath}")
            else:
                if not self._settings.unzip:
                    logger.warning(f"Expected file does not exist at {self.filepath}")


    @staticmethod
    def calculate_md5(file_path: Path, block_size: int = 2**16) -> str:
        md5 = hashlib.md5()
        with open(file_path, "rb") as f:
            for block in iter(lambda: f.read(block_size), b""):
                md5.update(block)
        return md5.hexdigest()

    @abstractmethod
    def create_dataset(self, *args, **kwargs) -> Mapping[str, DatasetProtocol]:
        raise NotImplementedError

    def create_datastreamer(
        self, batchsize: int, **kwargs
    ) -> Mapping[str, DatastreamerProtocol]:
        datasets = self.create_dataset()
        traindataset = datasets["train"]
        validdataset = datasets["valid"]
        # get preprocessor from kwargs
        preprocessor: Optional[Callable] = kwargs.pop("preprocessor", None)

        trainstreamer = BaseDatastreamer(
            traindataset, batchsize=batchsize, preprocessor=preprocessor
        )
        validstreamer = BaseDatastreamer(
            validdataset, batchsize=batchsize, preprocessor=preprocessor
        )
        return {"train": trainstreamer, "valid": validstreamer}


class BaseDatastreamer:
    """This datastreamer wil never stop
    The dataset should have a:
        __len__ method
        __getitem__ method

    """

    def __init__(
        self,
        dataset: DatasetProtocol,
        batchsize: int,
        preprocessor: Optional[Callable] = None,
    ) -> None:
        self.dataset = dataset
        self.batchsize = batchsize

        if preprocessor is None:
            self.preprocessor = lambda x: zip(*x)
        else:
            self.preprocessor = preprocessor

        self.size = len(self.dataset)
        self.reset_index()

    def __len__(self) -> int:
        return int(len(self.dataset) / self.batchsize)

    def __repr__(self) -> str:
        return f"BasetDatastreamer: {self.dataset} (streamerlen {len(self)})"

    def reset_index(self) -> None:
        self.index_list = np.random.permutation(self.size)
        self.index = 0

    def batchloop(self) -> Sequence[Tuple]:
        batch = []
        for _ in range(self.batchsize):
            x, y = self.dataset[int(self.index_list[self.index])]
            batch.append((x, y))
            self.index += 1
        return batch

    def stream(self) -> Iterator:
        while True:
            if self.index > (self.size - self.batchsize):
                self.reset_index()
            batch = self.batchloop()
            X, Y = self.preprocessor(batch)  # noqa N806
            yield X, Y



import re
import shutil
from typing import Any, Mapping

import pandas as pd
import polars as pl
from loguru import logger

from mads_datasets.base import (
    AbstractDatasetFactory,
    DatasetProtocol,
    DatastreamerProtocol,
)
from mads_datasets.datasets import PdDataset, PolarsDataset, TextDataset
from mads_datasets.datatools import keep_subdirs_only, walk_dir
from mads_datasets.settings import (
    DatasetSettings,
    PdDatasetSettings,
    SecureDatasetSettings,
    TextDatasetSettings,
)


class PenguinsDatasetFactory(AbstractDatasetFactory[DatasetSettings]):
    def create_dataset(self, *args, **kwargs) -> Mapping[str, DatasetProtocol]:
        pass


class IrisDatasetFactory(AbstractDatasetFactory[PdDatasetSettings]):
    def create_dataset(
        self, *args: Any, **kwargs: Any
    ) -> Mapping[str, DatasetProtocol]:
        self.download_data()
        df = pd.read_csv(
            self.filepath,
            header=None,
            names=self._settings.features + [self._settings.target],
        )
        df["sepal_width"].fillna(value=df["sepal_width"].mean(), inplace=True)
        df[self._settings.target] = df[self._settings.target].apply(
            lambda x: re.sub(r"setsoa", "setosa", x)
        )
        df.drop_duplicates(inplace=True)
        df["petal_width"] = df["petal_width"].apply(
            lambda x: float(re.findall(r"(\d+\.?\d*) mm", x)[0]) / 10
        )
        idx = df[df["sepal_length"] == 58].index
        df.loc[idx, "sepal_length"] = 5.8

        train = df.sample(frac=0.6)
        valid = df.drop(train.index)

        trainset = PdDataset(
            train, features=self._settings.features, target=self._settings.target
        )
        validset = PdDataset(
            valid, features=self._settings.features, target=self._settings.target
        )
        return {"train": trainset, "valid": validset}

    def create_datastreamer(
        self, batchsize: int, **kwargs
    ) -> Mapping[str, DatastreamerProtocol]:
        raise NotImplementedError("Not implemented yet...")


class IMDBDatasetFactory(AbstractDatasetFactory[TextDatasetSettings]):
    def create_dataset(
        self, *args: Any, **kwargs: Any
    ) -> Mapping[str, DatasetProtocol]:
        self.download_data()
        testdir = self.subfolder / "aclImdb/test"
        traindir = self.subfolder / "aclImdb/train"
        keep_subdirs_only(testdir)
        keep_subdirs_only(traindir)

        # remove dir with unlabeled reviews
        unsup = traindir / "unsup"
        if unsup.exists():
            shutil.rmtree(traindir / "unsup")

        formats = [f.value for f in self._settings.formats]
        trainpaths = [path for path in walk_dir(traindir) if path.suffix in formats]
        testpaths = [path for path in walk_dir(testdir) if path.suffix in formats]
        logger.info(
            f"Creating TextDatasets from {len(trainpaths)} trainfiles"
            f"and {len(testpaths)} testfiles."
        )

        traindataset = TextDataset(paths=trainpaths)
        testdataset = TextDataset(paths=testpaths)
        return {"train": traindataset, "valid": testdataset}


class SecureDatasetFactory(AbstractDatasetFactory[SecureDatasetSettings]):
    def create_dataset(
        self, *args: Any, **kwargs: Any
    ) -> Mapping[str, DatasetProtocol]:
        self.download_data()

        df = pl.read_parquet(self.filepath)
        dataset = PolarsDataset(df)

        return {"train": dataset}

class FavoritaDatasetFactory(AbstractDatasetFactory[DatasetSettings]):
    def create_dataset(
        self, *args: Any, **kwargs: Any
    ) -> Mapping[str, DatasetProtocol]:
        self.download_data()

        trainfile = self.subfolder / "train.parq"

        df = pl.read_parquet(trainfile)
        dataset = PolarsDataset(df)

        return {"train": dataset}
